"""Common functionality for REST API

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

import logging
from enum import Enum

from fastapi import Query
from pydantic import BaseModel
from pydantic import Field
from qamlib import V4L2Exception

from quark_backend.quark_camera import DEF_DEVICE
from quark_backend.utils import logger_conf


API_ROUTE = "/api"
V4L2_ROUTE = API_ROUTE + "/v4l2"

app_logger = logger_conf("fast_api_app", c_level=logging.DEBUG)


class Tags(str, Enum):
    """Available API tags"""

    v4l2_controls = "V4L2 Controls"
    """: V4l2 control related functions"""
    hl_controls = "Controls"
    """: High-level control related functions"""
    frame_parameters = "Frame Parameters"
    """: V4l2 frame parameters related functions"""
    streaming = "Streaming"
    """: Image streaming related functions"""
    image = "Images"
    """: Single image related functions"""
    devices = "Devices"
    """: Device discovery functions"""
    oculus = "Oculus"
    """: Oculus HSI related functions"""


class V4L2Endpoints(str, Enum):
    """Available V4L2 related API endpoints"""

    control = "/control"
    """: Functions operating on single v4l2 controls"""
    controls = "/controls"
    """: Functions operating on lists of v4l2 controls"""
    resolution = "/resolution"
    """: Functions controlling the frame resolution"""
    crop = "/crop"
    """: Functions controlling frame cropping"""
    framerate = "/framerate"
    """: Functions controlling the framerate"""
    pixel_format = "/pixel_format"
    """: Functions controlling the pixel format of the frames"""
    stream = "/stream"
    """: Functions controlling the streaming state"""


class APIError(Exception):
    """All custom API Exceptions"""

    pass


class APIV4L2Error(APIError):
    """Custom V4L2 Error Class."""

    code = 422
    description = "V4L2 Error"


class APIInputError(APIError):
    """Custom Input Error Class."""

    code = 422
    description = "API Input Error"


class APIGenError(APIError):
    """Generic Server Error Class."""

    code = 500
    description = "Internal Server Error"


def check_exceptions(func, *args):
    """Check for exceptions from Camera functions"""
    try:
        res = func(*args)
    except V4L2Exception as e:
        raise APIV4L2Error(str(e)) from e
    except RuntimeError as e:
        raise APIInputError(str(e)) from e
    except Exception as e:
        raise APIGenError(f"{type(e).__name__} from {func.__name__}(): {e}") from e

    return res


class CommonParams(BaseModel):
    """Parameters common between endpoints"""

    dev: str = Field(Query(default=DEF_DEVICE, description="target video device"))
    """: target video device"""
