"""REST API interface for high-level controls.

:Authors:
    Anton Mølbjerg Eskildsen <anton@qtec.com>
"""

from __future__ import annotations

from typing import Annotated
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Union

import numpy as np
import orjson
from fastapi import APIRouter
from fastapi import Body
from fastapi import Depends
from fastapi import Path
from pydantic import BaseModel
from pydantic import Field
from pydantic import RootModel
from qamlib import V4L2Exception
from quarklib.controls import DriverControl
from quarklib.controls import GroupControl
from quarklib.controls import IControl
from quarklib.controls import formatters
from starlette.responses import JSONResponse

from quark_backend.api.common import API_ROUTE
from quark_backend.api.common import APIGenError
from quark_backend.api.common import APIInputError
from quark_backend.api.common import APIV4L2Error
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.api.common import check_exceptions
from quark_backend.api.examples import control_value_examples
from quark_backend.api.examples import list_controls_responses
from quark_backend.api.examples import path_examples
from quark_backend.api.examples import read_control_responses
from quark_backend.quark_camera import Camera


router = APIRouter(prefix=API_ROUTE, tags=[Tags.hl_controls])


Value = float | int | list | str

T = TypeVar("T", float, int, str)
U = TypeVar("U", float, int, str, list[int | float], list[list[int | float]])


def recursive_arrays_to_lists(val):
    if isinstance(val, dict):
        for k, v in val.items():
            if isinstance(v, np.ndarray):
                val[k] = v.tolist()
            else:
                recursive_arrays_to_lists(v)


class ArrayValue(RootModel):
    root: list[int | float] | list[list[int | float]]


class ControlValue(RootModel):
    root: bool | float | int | str | ArrayValue


# class GroupControlValue(BaseModel):
#     children: list[ControlValueT]


class GroupValue(RootModel):
    root: dict[str, ControlValue | GroupValue]


class Interval(BaseModel, Generic[T]):
    min: T
    max: T
    step: T


class ControlFormat(BaseModel, Generic[T]):
    type: str
    range: Interval[float | int] | set[int | str]


ControlInfoT = Annotated[
    Union["ControlInfo", "GroupInfo"], Field(union_mode="left_to_right")
]


class ControlInfoBase(BaseModel):
    name: str = Field(description="V4L2 name of the control.")
    path: str = Field(description="Control path. Use when reading/writing controls.")
    description: str


class ControlInfo(ControlInfoBase):
    format: ControlFormat = Field(
        description="Description of the control type and valid values."
    )


get_controls_example = ["red balance", "green balance", "blue balance"]


class GroupInfo(ControlInfoBase):
    children: list[ControlInfoT] = Field(
        description="Children are accessed via the group's path."
    )


# class ControlListResponse(BaseModel):
#     root: ControlInfoT


def create_info_model(ctl: IControl, path: list[str]) -> ControlInfoT:
    new_path = path + [ctl.name]
    if isinstance(ctl, GroupControl):
        children = [create_info_model(v, new_path) for v in ctl.children]
        return GroupInfo(
            name=ctl.name,
            path="/" + "/".join(new_path),
            description=ctl.__doc__,
            children=children,
        )
    elif isinstance(ctl, DriverControl):
        rn = ctl.formatter.valid_set
        if isinstance(rn, formatters.Interval):
            rn = Interval(min=rn.lower, max=rn.upper, step=rn.step)
        return ControlInfo(
            name=ctl.name,
            path="/" + "/".join(new_path),
            description=ctl.__doc__,
            format=ControlFormat(type=ctl.formatter.__class__.__name__, range=rn),
        )
    else:
        raise ValueError(f"Unreachable: {ctl.__class__.__name__}")


def create_info_list(ctl: IControl) -> str | dict:
    if isinstance(ctl, GroupControl):
        children = [create_info_list(v) for v in ctl.children]
        return {
            "name": ctl.name,
            "children": children,
        }
    elif isinstance(ctl, DriverControl):
        return ctl.name
    else:
        raise ValueError(f"Unreachable: {ctl.__class__.__name__}")


def get_recursive(path: list[str], ctl):
    if len(path) == 0:
        return ctl
    else:
        return get_recursive(path[1:], getattr(ctl, path[0]))


class OrjsonResponse(JSONResponse):
    def render(self, content: Any) -> bytes:
        return orjson.dumps(content, option=orjson.OPT_SERIALIZE_NUMPY)


class GroupControlDef(BaseModel):
    name: str
    children: list[str | GroupControlDef]


@router.get("/controls/list", responses=list_controls_responses)
def list_controls(params: CommonParams = Depends()) -> list[GroupControlDef | str]:
    """
    List controls.

    The response body represents the control hierarchy and is useful for
    determining control paths. For example, in the provided example response,
    the path of *red* is *digital_gain/red*.
    ```
    """

    def inner():
        ctl = Camera._get_camera(params.dev).controls
        return [create_info_list(c) for c in ctl.children]

    return check_exceptions(inner)


def check_exceptions_ctl(func, *args):
    """Check exceptions including missing controls."""
    try:
        res = func(*args)
    except V4L2Exception as e:
        raise APIV4L2Error(str(e)) from e
    except AttributeError:
        raise APIInputError("Control not found")
    except RuntimeError as e:
        raise APIInputError(str(e)) from e
    except Exception as e:
        raise APIGenError(f"{type(e).__name__} from {func.__name__}(): {e}") from e

    return res


@router.get("/controls/info/{path:path}")
def get_controls_info(
    path: str = Path(openapi_examples=path_examples), params: CommonParams = Depends()
) -> ControlInfoT:
    """
    Get detailed control info.

    The response contains detailed information about the control(s) requested.
    """

    def f():
        if path == "" or path == "root":
            ctl = Camera._get_camera(params.dev).controls
            ctl.name = ""
            ml = [create_info_model(c, []) for c in ctl.children]
            m = GroupInfo(name="root", path="/", description=ctl.__doc__, children=ml)
        else:
            path_list = path.split("/")
            ctl = get_recursive(path_list, Camera._get_camera(params.dev).controls)
            m = create_info_model(ctl, path_list[:-1])
        return m

    return check_exceptions_ctl(f)


@router.get(
    "/controls/{path:path}",
    response_model=ControlValue | GroupValue,
    responses=read_control_responses,
)
def read_control(
    path: str = Path(
        description="Specifies the control/group to read. Use *list_controls* to discover control paths.",
        openapi_examples=path_examples,
    ),
    params: CommonParams = Depends(),
) -> OrjsonResponse:
    """
    Read the value of a single control or a group of controls.
    """

    if path == "" or path == "all":
        path = []
    else:
        path = path.split("/")

    def f():
        ctl = get_recursive(path, Camera._get_camera(params.dev).controls)
        return OrjsonResponse(content=ctl.read())

    return check_exceptions_ctl(f)


@router.post(
    "/controls/{path:path}",
    response_model=ControlValue | GroupValue,
    responses=read_control_responses,
)
def write_control(
    path: str = Path(
        description="Specifies which control/group to write to.",
        openapi_examples={
            "red_gain": {"summary": "Digital gain red", "value": "digital_gain/red"},
            "ccm": {
                "summary": "Color conversion group",
                "value": "color_conversion",
            },
            "ccm_offset": {
                "summary": "Color conversion offset vector",
                "value": "color_conversion/ccm/offset",
            },
        },
    ),
    value: ControlValue | GroupValue = Body(
        description="The value to write. Use `controls/info/{path}` to discover the formats for different controls.",
        openapi_examples=control_value_examples,
    ),
    params: CommonParams = Depends(),
) -> OrjsonResponse:
    """
    Write the value of a single control or a group of controls.

    The path is the position of the control in the hierarchy.
    Use *list_controls* to discover control paths.

    """

    path = path.split("/")

    def f():
        ctl = get_recursive(path, Camera._get_camera(params.dev).controls)
        print(value.model_dump())
        ctl.write(value.model_dump())
        return OrjsonResponse(content=ctl.read())

    return check_exceptions_ctl(f)
