"""REST API video device interface

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

import os
from typing import List

from fastapi import APIRouter
from fastapi import Depends
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from pydantic import Field

from ..quark_camera import Camera
from .common import API_ROUTE
from .common import CommonParams
from .common import Tags
from .common import check_exceptions


router = APIRouter(prefix=API_ROUTE, tags=[Tags.devices])


class DeviceList(BaseModel):
    """List of video devices in the camera"""

    video: List[str] = Field(
        description="generic video devices",
        example=[
            "/dev/video5",
            "/dev/video4",
            "/dev/video3",
            "/dev/video2",
            "/dev/video1",
            "/dev/video0",
        ],
    )
    """: generic video devices"""
    qtec_video: List[str] = Field(
        description="qtec video capture devices", example=["/dev/qtec/video0"]
    )
    """: qtec video capture devices"""
    qtec: List[str] = Field(
        description="qtec video devices",
        example=[
            "/dev/qtec/xform_dist_readback0",
            "/dev/qtec/xform_gain0",
            "/dev/qtec/xform_dist0",
            "/dev/qtec/video0",
            "/dev/qtec/xform_gain_readback0",
        ],
    )
    """: qtec video devices"""


def get_video_device_nrs(path: str = "/dev", name: str = "video") -> List[int]:
    """List video devices numbers

    Filter based on specified name pattern

    :param path: path to look for video devices
    :param name: name pattern used to filter video devices
    :return: list of video device numbers
    """
    devs = os.listdir(path)
    vid_indices = [int(dev[-1]) for dev in devs if dev.startswith(name)]
    vid_indices = sorted(vid_indices)
    return vid_indices


def list_video_devices(path: str = "/dev", name: str = "video") -> List[str]:
    """List video devices with specified name pattern

    :param path: path to look for video devices
    :param name: name pattern used to filter video devices
    :return: list of video device names (full path)
    """
    devs = os.listdir(path)
    devices = [f"{path}/{dev}" for dev in devs if dev.startswith(name)]
    return devices


def list_qtec_video_devices() -> List[str]:
    """List qtec video capture devices"""
    return list_video_devices("/dev/qtec", name="video")


def list_qtec_devices() -> List[str]:
    """List qtec video devices"""
    return list_video_devices("/dev/qtec", name="")


@router.get(
    "/list_devices",
    response_model=DeviceList,
    summary="List video devices",
    response_description="Lists of different video devices",
)
def list_devices() -> JSONResponse:
    """List video devices (both generic and qtec)"""
    return DeviceList(
        video=list_video_devices(),
        qtec_video=list_qtec_video_devices(),
        qtec=list_qtec_devices(),
    )


@router.get("/stop")
def stop_stream(params: CommonParams = Depends()):
    check_exceptions(Camera.stop, params.dev)
