"""REST API single image interface

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

import sys
from enum import Enum

import numpy as np
from fastapi import APIRouter
from fastapi import Depends
from fastapi.responses import Response
from qamlib import V4L2Exception

from quark_backend.api.common import API_ROUTE
from quark_backend.api.common import APIGenError
from quark_backend.api.common import APIInputError
from quark_backend.api.common import APIV4L2Error
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.quark_camera import Camera


try:
    import cv2

    HAVE_OPENCV = True
except ImportError:
    HAVE_OPENCV = False


router = APIRouter(
    prefix=API_ROUTE,
    tags=[Tags.image],
)


def needs_byteswap(frame: np.ndarray):
    """Check if a byte-swap is needed for the PNM format."""
    # 8-bit
    if frame.itemsize == 1:
        return False
    # Big Endian
    if frame.dtype.byteorder == ">":
        return False
    # Big Endian native byte order
    if frame.dtype.byteorder == "=" and sys.byteorder == "big":
        return False
    return True


def make_pnm(frame: bytes, format: str) -> bytes:
    """PNM encode an image frame

    Note: the PNM expects the colors in RGB order

    :return: PNM encoded bytes
    """
    match format:
        case "RGB3" | "GREY" | "QG08":
            data = frame
        case "Y16" | "QG16" | "RGB6":
            if needs_byteswap(frame):
                data = frame.byteswap()
            else:
                data = frame
        case "RGB4":
            data = frame[:, :, 1:]
        case "BGR3":
            data = frame[..., ::-1].copy()
        case "BGR4":
            data = frame[:, :, :3]
            data = data[..., ::-1].copy()
        case "GBRG" | "GRBG" | "RGGB" | "BA81":
            if not HAVE_OPENCV:
                raise ValueError(
                    f"This format ({format}) requires OpenCV for conversion."
                )
        case _:
            raise ValueError(f"Unsupported image format {format}")

    match format:
        case "GBRG":
            data = cv2.cvtColor(frame, cv2.COLOR_BayerGR2RGB)
        case "GRBG":
            data = cv2.cvtColor(frame, cv2.COLOR_BayerGB2RGB)
        case "RGGB":
            data = cv2.cvtColor(frame, cv2.COLOR_BayerBG2RGB)
        case "BA81":
            data = cv2.cvtColor(frame, cv2.COLOR_BayerRG2RGB)

    height, width, channels = data.shape

    if channels == 1:
        type = "P5"
    else:
        type = "P6"

    if data.nbytes == width * height * channels:
        max_val = 255
    else:
        max_val = 65535

    pnm_header = f"{type} {width} {height} {max_val}\n"
    pnm = b""
    pnm += bytearray(pnm_header, "ascii")
    pnm += data.tobytes()
    return pnm


def make_pam(frame: bytes, format: str) -> bytes:
    """PAM encode an image frame

    :return: PAM encoded bytes
    """
    height, width, channels = frame.shape

    if frame.nbytes == width * height * channels:
        max_val = 255
    else:
        max_val = 65535

    if needs_byteswap(frame):
        data = frame.byteswap()
    else:
        data = frame

    header = "P7\n"
    header += f"WIDTH {width}\n"
    header += f"HEIGHT {height}\n"
    header += f"DEPTH {channels}\n"
    header += f"MAXVAL {max_val}\n"
    header += f"TUPLTYPE {format}\n"
    header += "ENDHDR\n"
    pam = b""
    pam += bytearray(header, "ascii")
    pam += data.tobytes()
    return pam


class ImageEncodings(str, Enum):
    """Available image encodings"""

    raw = "RAW"
    """: No encoding"""
    jpeg = "JPEG"
    """: JPEG encoding"""
    pnm = "PNM"
    """: PNM encoding"""
    pam = "PAM"
    """: PAM encoding"""


@router.get(
    "/get_frame/{encoding}",
    responses={
        200: {"content": {"image/jpeg": {}, "application/octet-stream": {}}},
    },
    response_class=Response,
)
def get_image(
    encoding: ImageEncodings = ImageEncodings.jpeg,
    params: CommonParams = Depends(),  # noqa: B008
) -> Response:
    """
    Single image route.

    Put this in the src attribute of an img tag if JPEG.
    """
    try:
        camera = Camera(device=params.dev, stream=True)

        force_download = False
        filename = "img"

        if encoding is ImageEncodings.jpeg:
            img = camera.get_frame_jpeg().data
            mimetype = "image/jpeg"
        elif (
            encoding is ImageEncodings.raw
            or encoding is ImageEncodings.pnm
            or encoding is ImageEncodings.pam
        ):
            mimetype = "application/octet-stream"
            force_download = True
            frame = camera.get_frame()
            if encoding is ImageEncodings.pnm:
                img = make_pnm(frame.data, frame.pix_fmt)
                filename = "img.pnm"
            elif encoding is ImageEncodings.pam:
                img = make_pam(frame.data, frame.pix_fmt)
                filename = "img.pam"
            else:
                img = frame.data.tobytes()
                filename = "img.raw"

        resp = Response(img, media_type=mimetype)
        if force_download:
            resp.headers["Content-Disposition"] = f"attachment; filename={filename}"
        return resp

    except V4L2Exception as e:
        raise APIV4L2Error(str(e)) from e
    except RuntimeError as e:
        raise APIInputError(str(e)) from e
    except Exception as e:
        raise APIGenError(
            f"{type(e).__name__} from {sys._getframe().f_code.co_name}(): {e}"
        ) from e
