import numpy as np
from fastapi import APIRouter
from fastapi import Body
from pydantic import BaseModel
from pydantic import Field

from quark_backend.api.common import API_ROUTE
from quark_backend.api.common import Tags
from quark_backend.api.common import check_exceptions
from quark_backend.oculus import HorizontalCrop
from quark_backend.oculus import OculusConfig
from quark_backend.oculus import OculusManager
from quark_backend.oculus import OculusSettings
from quark_backend.oculus import OculusSettingsRequest
from quark_backend.oculus import SpectralBandSelection
from quark_backend.quark_camera import Camera


router = APIRouter(
    prefix=API_ROUTE + "/oculus",
    tags=[Tags.oculus],
)


@router.get("/config")
def get_config() -> OculusConfig:
    """Read current Oculus configuration."""
    return check_exceptions(lambda: OculusManager.get_config().config)


@router.get("/settings")
def get_settings() -> OculusSettings:
    """Read Oculus settings."""
    return check_exceptions(lambda: OculusManager.get_config().settings)


@router.post("/settings")
def set_settings(
    settings: OculusSettingsRequest = Body(),
) -> OculusSettings:
    """Modify Oculus settings."""

    def f():
        cfg = OculusManager.get_config()
        if settings.horizontal_crop is not None:
            cfg.settings.horizontal_crop = settings.horizontal_crop
        if settings.bands is not None:
            cfg.settings.bands = settings.bands
        if settings.exposure is not None:
            cfg.settings.exposure = settings.exposure
        if settings.vertical_binning is not None:
            cfg.settings.vertical_binning = settings.vertical_binning
        if settings.horizontal_binning is not None:
            cfg.settings.horizontal_binning = settings.horizontal_binning

        cfg.apply_settings()
        OculusManager.save_settings(cfg.settings)
        return OculusManager().get_config().settings

    return check_exceptions(f)


@router.get("/bands")
def get_bands() -> SpectralBandSelection:
    """Get current band selection."""
    return check_exceptions(lambda: OculusManager.get_config().settings.bands)


wavelength_responses = {
    200: {
        "description": "Success",
        "content": {
            "application/json": {
                "examples": {
                    "full": {
                        "summary": "GSENSE full range",
                        "value": list(np.linspace(430, 1000, 350)),
                    }
                },
            }
        },
    },
}


class CropDimensionResponse(BaseModel):
    max_width: int
    max_height: int
    current_width: int
    current_height: int


@router.get("/crop_dimensions")
def get_crop_dimensions() -> CropDimensionResponse:
    def f():
        c = Camera._get_camera().get_crop()[0]
        w, h = OculusManager.get_config().get_max_crop()
        return CropDimensionResponse(
            max_width=w, max_height=h, current_width=c.width, current_height=c.height
        )

    return check_exceptions(f)


@router.get("/wavelengths", responses=wavelength_responses)
def get_wavelengths() -> list[float]:
    """Get wavelengths in nm of each band in the current selection."""
    return check_exceptions(lambda: list(OculusManager.get_config().wavelengths()))


spectral_crop_examples = {
    "single_range": {
        "summary": "Single range",
        "description": "Select all bands in the specified range (min, max)",
        "value": [[0, 350]],
    },
    "single_range_steps": {
        "summary": "Stepped range",
        "description": "Select each n'th band in the specified range. The third number defines the step size.",
        "value": [[0, 350, 5]],
    },
    "single_band": {
        "summary": "Single band",
        "description": "Select a single band",
        "value": [50],
    },
    "multi": {
        "summary": "Multiple selections",
        "description": "Selection types can be combined to select arbitrary ranges and individual bands. Note that each band can be selected at most once.",
        "value": [[0, 20], [40, 100, 2], 120, 230, [250, 255]],
    },
}


@router.post("/bands")
def set_bands(
    bands: SpectralBandSelection = Body(openapi_examples=spectral_crop_examples),
) -> SpectralBandSelection:
    """Update band selection."""

    def f():
        cfg = OculusManager.get_config()
        cfg.settings.bands = bands
        cfg.apply_settings()
        OculusManager.save_settings(cfg.settings)
        return bands

    return f()


@router.post("/horizontal_crop")
def set_horizontal_crop(
    crop: HorizontalCrop = Body(
        description="Start and end x positions of crop",
        example=HorizontalCrop(start=50, end=1700),
    ),
):
    def f():
        cfg = OculusManager.get_config()
        cfg.settings.horizontal_crop = crop
        cfg.apply_settings()
        OculusManager.save_settings(cfg.settings)
        return OculusManager.get_config().settings.horizontal_crop

    return check_exceptions(f)


points_x = np.array([0, 330, 400]).reshape(-1, 1)
points_y = np.array([800, 500, 550])
A = np.hstack([points_x**2, points_x, np.ones((3, 1))])
param = np.linalg.solve(A, points_y)
x = np.arange(350)
y = (param[0] * x**2 + param[1] * x + param[2]).astype(np.uint32)
set_exposure_examples = {
    "single": {
        "summary": "Setting exposure globally",
        "description": "Applies the exposure to all bands.",
        "value": 500,
    },
    "list": {
        "summary": "Set exposure for each band in crop",
        "description": """Use individual exposure values for each band.

This option is only supported on GSENSE sensors.

This option is useful for correcting non-linearities in the sensor response
and light source. The example uses a quadratic curve through the points
(0, 800), (330, 500), (400, 550) to approximate the non-linear light emission of
a halogen bulb.""",
        "value": y.tolist(),
    },
}


class BinningRequest(BaseModel):
    horizontal: int | None = Field(default=None, example=2, le=6, gt=0)
    vertical: int | None = Field(default=None, example=2, le=6, gt=0)


@router.post("/binning")
def set_binning(body: BinningRequest = Body()) -> BinningRequest:
    cfg = OculusManager.get_config()
    settings = cfg.settings

    if body.horizontal is not None:
        # ratio = settings.horizontal_binning / body.horizontal
        # settings.horizontal_crop = HorizontalCrop(
        #     start=int(settings.horizontal_crop.start * ratio),
        #     end=int(settings.horizontal_crop.end * ratio),
        # )

        settings.horizontal_binning = body.horizontal

    if body.vertical is not None:
        bands = settings.bands.root
        ratio = settings.vertical_binning / body.vertical
        for i in range(len(bands)):
            if isinstance(bands[i], int):
                bands[i] = int(bands[i] * ratio)
            if isinstance(bands[i], tuple):
                if len(bands[i]) == 2:
                    bands[i] = (
                        int(bands[i][0] * ratio),
                        int(bands[i][1] * ratio),
                    )
                else:
                    bands[i] = (
                        int(bands[i][0] * ratio),
                        int(bands[i][1] * ratio),
                        max(1, int(bands[i][2] * ratio)),
                    )

        print("bb", bands)
        settings.bands = SpectralBandSelection(bands)
        settings.vertical_binning = body.vertical

    cfg.apply_settings()
    OculusManager.save_settings(settings)

    return BinningRequest(
        horizontal=settings.horizontal_binning, vertical=settings.vertical_binning
    )


@router.post("/exposure")
def set_exposure(
    exposure: int | list[int] = Body(
        description="Exposure in microseconds s", openapi_examples=set_exposure_examples
    ),
) -> int | list[int]:
    """Set frame exposure in microseconds."""

    def f():
        cfg = OculusManager.get_config()
        cfg.settings.exposure = exposure
        cfg.apply_settings()
        OculusManager.save_settings(cfg.settings)
        return OculusManager.get_config().settings.exposure

    return check_exceptions(f)


@router.post("/config")
def write_config(config: OculusConfig = Body()):
    def f():
        cfg = OculusManager.get_config()
        OculusManager.save_config(config)

        cfg.apply_settings()

        return cfg.config

    return f()


@router.post("/default_settings")
def set_default_settings():
    def f():
        cfg = OculusManager.get_config()
        cfg.settings.horizontal_crop = HorizontalCrop(
            start=0, end=cfg.config.spectrograph.sensor_crop_width
        )
        cfg.settings.horizontal_binning = 1
        cfg.settings.vertical_binning = 1
        cfg.settings.bands.root = [(0, cfg.config.spectrograph.sensor_crop_height)]
        cfg.apply_settings()
        OculusManager.save_settings(cfg.settings)
        return cfg.settings

    return check_exceptions(f)
