"""REST API streaming interface

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

import asyncio
import sys
import time
from typing import Generator

from fastapi import APIRouter
from fastapi import Depends
from fastapi import Request
from fastapi.responses import StreamingResponse
from qamlib import V4L2Exception

from quark_backend.api.common import APIGenError
from quark_backend.api.common import APIInputError
from quark_backend.api.common import APIV4L2Error
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.api.common import app_logger
from quark_backend.base_camera import BaseCamera
from quark_backend.quark_camera import Camera
from quark_backend.tcp_socket import TCPSocket


# from fastapi import HTTPException

# from pydantic import create_model_from_typeddict
# https://pydantic-docs.helpmanual.io/usage/models/


router = APIRouter(tags=[Tags.streaming])


async def async_gen(camera: BaseCamera) -> Generator[bytes, None, None]:
    """Video streaming async generator function."""
    yield b"--frame\r\n"
    retry = False
    retries = 5
    while True:
        try:
            if retry:
                if retries <= 0:
                    app_logger.error("No more retries")
                    break

                app_logger.info(f"Retrying streaming in 1 second: {retries} left")
                retries -= 1
                await asyncio.sleep(1)
                camera = Camera(device=camera.video_device, stream=True)

            jpeg = camera.get_frame_jpeg()
            if jpeg is not None:
                retry = False
                retries = 5
                yield TCPSocket.package_mjpeg(jpeg.data, jpeg.metadata)
            else:
                # raise V4L2Exception("No jpeg frame available")
                app_logger.warning(
                    f"No jpeg frame available from: {camera.video_device}"
                )
                retry = True

            # need this sleep in order to be able to catch the disconnect
            await asyncio.sleep(0)

        except asyncio.CancelledError:
            app_logger.debug("Client disconnected")
            break
        except V4L2Exception as e:
            app_logger.debug(f"V4L2Exception in async_gen: {e}")
            retry = True
            continue
        except Exception as e:
            app_logger.error(f"Unhandled exception in async_gen: {e}")
            break


def sync_gen(camera: BaseCamera) -> Generator[bytes, None, None]:
    """Video streaming generator function."""
    yield b"--frame\r\n"
    retry = False
    retries = 5
    while True:
        try:
            if retry:
                if retries <= 0:
                    app_logger.error("No more retries")
                    break

                app_logger.info(f"Retrying streaming in 1 second: {retries} left")
                retries -= 1
                time.sleep(1)
                camera = Camera(device=camera.video_device, stream=True)

            jpeg = camera.get_frame_jpeg()
            if jpeg is not None:
                retry = False
                retries = 5
                yield TCPSocket.package_mjpeg(jpeg.data, jpeg.metadata)
            else:
                # raise V4L2Exception("No jpeg frame available")
                app_logger.warning(
                    f"No jpeg frame available from: {camera.video_device}"
                )
                retry = True

        except V4L2Exception as e:
            app_logger.debug(f"V4L2Exception in sync_gen: {e}")
            retry = True
            continue
        except Exception as e:
            # FIXME: this is a temporary workaround to recover from qamlib
            # exceptions caused by changing resolution/crop while streaming
            # See https://gitlab.com/qtec/public/qamlib/-/issues/66
            app_logger.warning(f"Unhandled exception in sync_gen: {e}")
            retry = True
            continue


# http://10.10.150.225:5000/mjpeg?dev=/dev/qtec/video0


class MJPEGResponse(StreamingResponse):
    """Custom response for MJPEG stream"""

    media_type = "multipart/x-mixed-replace; boundary=frame"
    """: The returned media type"""


@router.get("/mjpeg/async", response_class=MJPEGResponse)
async def async_video_feed(
    request: Request,
    params: CommonParams = Depends(),  # noqa: B008
) -> MJPEGResponse:
    """Async video streaming route.

    Put this in the src attribute of an img tag.
    """
    app_logger.debug(f"mjpeg/async: {request.client}")
    # TODO: Use information from request.client (ip+port) in order to
    # identify the different clients since they share the same thread
    # and therefore share also the same frame queue
    try:
        cam = Camera(device=params.dev, stream=True)
        # get 1st frame to check if device is busy
        if cam.get_frame_jpeg() is None:
            raise APIV4L2Error(f"No frame available from {cam.video_device}")
        return MJPEGResponse(async_gen(cam))
    except V4L2Exception as e:
        raise APIV4L2Error(str(e)) from e
    except RuntimeError as e:
        raise APIInputError(str(e)) from e
    except Exception as e:
        raise APIGenError(
            f"{type(e).__name__} from {sys._getframe().f_code.co_name}(): {e}"
        ) from e
    # raise HTTPException(status_code=404, detail="Some error")


@router.get("/mjpeg", response_class=MJPEGResponse)
def video_feed(params: CommonParams = Depends()) -> MJPEGResponse:  # noqa: B008
    """Video streaming route.

    Put this in the src attribute of an img tag.
    """
    try:
        cam = Camera(device=params.dev, stream=True)
        # get 1st frame to check if device is busy
        if cam.get_frame_jpeg() is None:
            raise APIV4L2Error(f"No frame available from {cam.video_device}")
        return MJPEGResponse(sync_gen(cam))
    except V4L2Exception as e:
        raise APIV4L2Error(str(e)) from e
    except RuntimeError as e:
        raise APIInputError(str(e)) from e
    except Exception as e:
        raise APIGenError(
            f"{type(e).__name__} from {sys._getframe().f_code.co_name}(): {e}"
        ) from e
