"""REST API interface for single v4l2 control

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

from typing import Union

from fastapi import APIRouter
from fastapi import Body
from fastapi import Depends
from fastapi import Path

from quark_backend.api.common import V4L2_ROUTE
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.api.common import V4L2Endpoints
from quark_backend.api.common import check_exceptions
from quark_backend.quark_camera import Camera
from quark_backend.quark_camera import V4L2ControlValue


router = APIRouter(
    prefix=V4L2_ROUTE + V4L2Endpoints.control,
    tags=[Tags.v4l2_controls],
)


class ControlExamples:
    """Examples for controls API"""

    name = {
        "exposure": {
            "summary": "Exposure Time",
            "value": "Exposure Time, Absolute",
        },
        "gain": {"summary": "Gain", "value": "gain"},
        "red": {"summary": "Red Gain", "value": "red_balance"},
        "green": {"summary": "Green Gain", "value": "green_balance"},
        "blue": {"summary": "Blue Gain", "value": "blue_balance"},
        "fpn": {
            "summary": "Fixed Pattern Noise Correction",
            "value": "fixed_pattern_noise_correction",
        },
    }

    value = {
        "exposure": {
            "summary": "Exposure Time",
            "value": {"value": 5000},
        },
        "gain": {"summary": "Gain", "value": {"value": 0}},
        "red": {"summary": "Red Gain", "value": {"value": 16384}},
        "green": {"summary": "Green Gain", "value": {"value": 16384}},
        "blue": {"summary": "Blue Gain", "value": {"value": 16384}},
        "array": {"summary": "Array Control", "value": {"value": [1, 2, 3]}},
    }


get_control_responses = {
    200: {
        "description": "Success",
        "content": {
            "application/json": {
                "examples": {
                    **ControlExamples.value,
                    "sensor type": {
                        "summary": "Sensor Type",
                        "value": {"value": "CMV4000v3 Bayer"},
                    },
                }
            }
        },
    },
}


@router.get(
    "/{ctrl_id_or_name}",
    responses=get_control_responses,
)
def get_control(
    ctrl_id_or_name: Union[int, str] = Path(  # noqa: B008
        description="Id or Name of v4l2 control",
        openapi_examples={
            **ControlExamples.name,
            "sensor type": {
                "summary": "Sensor Type",
                "value": "Sensor Type",
            },
        },
    ),
    params: CommonParams = Depends(),  # noqa: B008
) -> V4L2ControlValue:
    """
    Get value of single v4l2 control

    Supports both normal and extended controls
    """
    return check_exceptions(Camera.get_control, ctrl_id_or_name, params.dev)


set_control_responses = {
    200: {
        "description": "Success",
        "content": {"application/json": {"examples": ControlExamples.value}},
    },
}


@router.post(
    "/{ctrl_id_or_name}",
    responses=set_control_responses,
)
def set_control(
    ctrl_id_or_name: Union[int, str] = Path(  # noqa: B008
        description="Id or Name of v4l2 control",
        openapi_examples=ControlExamples.name,
    ),
    value: V4L2ControlValue = Body(  # noqa: B008
        openapi_examples=ControlExamples.value, descripton="v4l2 control value"
    ),
    params: CommonParams = Depends(),  # noqa: B008
) -> V4L2ControlValue:
    """
    Set value of single v4l2 control

    Supports both normal and extended controls
    """
    return check_exceptions(Camera.set_control, ctrl_id_or_name, value, params.dev)
