"""REST API interface for multiple v4l2 controls

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

from typing import Annotated
from typing import Dict

from fastapi import APIRouter
from fastapi import Body
from fastapi import Depends
from fastapi import Query
from pydantic import Field

from quark_backend.api.common import V4L2_ROUTE
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.api.common import V4L2Endpoints
from quark_backend.api.common import check_exceptions
from quark_backend.quark_camera import Camera
from quark_backend.quark_camera import V4L2Control
from quark_backend.quark_camera import V4L2ControlValue


router = APIRouter(
    prefix=V4L2_ROUTE + V4L2Endpoints.controls,
    tags=[Tags.v4l2_controls],
)

get_controls_example = ["red balance", "green balance", "blue balance"]

set_controls_example = {
    "red_balance": {"value": 16384},
    "green_balance": {"value": 16384},
    "blue_balance": {"value": 16384},
}

get_controls_responses = {
    200: {
        "description": "Success",
        "content": {
            "application/json": {
                "example": set_controls_example,
            }
        },
    },
}


ControlsResponse = Annotated[dict[str, V4L2ControlValue], Field()]


@router.get(
    "/",
    responses=get_controls_responses,
)
def get_controls(
    params: CommonParams = Depends(),  # noqa: B008
    ctrl_list: Annotated[
        list[str],
        Query(
            example=get_controls_example,
            description="List of names of v4l2 controls from which to retrieve values",
        ),
    ] = None,
) -> dict[str, V4L2ControlValue]:
    """
    Get the values of multiple v4l2 controls

    Supports both normal and extended controls

    All controls will be returned if a list is not passed by query
    """
    return check_exceptions(Camera.get_controls, ctrl_list, params.dev)


@router.post(
    "/",
    responses=get_controls_responses,
)
def set_controls(
    params: CommonParams = Depends(),  # noqa: B008
    ctrl_list: Dict[str, V4L2ControlValue] = Body(  # noqa: B008
        default=...,
        description="mapping of control names and their desired values",
        example=set_controls_example,
    ),
) -> dict[str, V4L2ControlValue]:
    """
    Set the values of multiple v4l2 controls

    Supports both normal and extended controls
    """
    return check_exceptions(Camera.set_controls, ctrl_list, params.dev)


@router.get(
    "/list",
)
def list_controls(params: CommonParams = Depends()) -> list[V4L2Control]:  # noqa: B008
    """List all v4l2 controls

    Returns a list of v4l2 controls with all their properties listed.

    Note: does not include the control current value.
    """
    return check_exceptions(Camera.list_controls, params.dev)
