"""REST API framerate interface

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

from fastapi import APIRouter
from fastapi import Depends

from quark_backend.api.common import V4L2_ROUTE
from quark_backend.api.common import CommonParams
from quark_backend.api.common import Tags
from quark_backend.api.common import V4L2Endpoints
from quark_backend.api.common import check_exceptions
from quark_backend.quark_camera import Camera
from quark_backend.quark_camera import ContinuousFrameRate
from quark_backend.quark_camera import DiscreteFrameRate
from quark_backend.quark_camera import FrameRate
from quark_backend.quark_camera import StepwiseFrameRate


router = APIRouter(
    prefix=V4L2_ROUTE + V4L2Endpoints.framerate,
    tags=[Tags.frame_parameters],
)


@router.get(
    "/",
    response_model=FrameRate,
)
def get_framerate(params: CommonParams = Depends()) -> FrameRate:  # noqa: B008
    """Get camera framerate"""
    return check_exceptions(Camera.get_framerate, params.dev)


list_framerates_response = {
    200: {
        "description": "Success",
        "content": {
            "application/json": {
                "examples": {
                    "discrete framerate": {
                        "summary": "Discrete Framerate",
                        "value": DiscreteFrameRate(values=[10, 20, 30]),
                    },
                    "continuous framerate": {
                        "summary": "Continuous Framerate",
                        "value": ContinuousFrameRate(min=1, max=30),
                    },
                    "stepwise framerate": {
                        "summary": "Stepwise Framerate",
                        "value": StepwiseFrameRate(min=1, max=30, step=1),
                    },
                }
            }
        },
    },
}


@router.get(
    "/list",
    responses=list_framerates_response,
)
def list_framerates(
    params: CommonParams = Depends(),
) -> DiscreteFrameRate | ContinuousFrameRate | StepwiseFrameRate:  # noqa: B008
    """
    List possible framerates for this camera with current settings

    Depends on the current pixel format and frame size
    """
    return check_exceptions(Camera.list_framerates, params.dev)


@router.post(
    "/",
    response_model=FrameRate,
)
def set_framerate(
    framerate: FrameRate,
    params: CommonParams = Depends(),  # noqa: B008
) -> FrameRate:
    """Set camera framerate"""
    return check_exceptions(Camera.set_framerate, framerate.value, params.dev)
