"""Base Camera"""

import logging
import threading
import time
from threading import get_ident
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple

import numpy as np
from qamlib import FrameMetadata
from qamlib import V4L2Exception
from turbojpeg import TurboJPEG, TJPF_RGB, TJPF_BGR, TJPF_ARGB, TJPF_BGRA, TJPF_GRAY

from quark_backend.utils import DeferredImport
from quark_backend.utils import get_video_device
from quark_backend.utils import get_video_device_str
from quark_backend.utils import logger_conf


cv2 = DeferredImport("cv2")

DEF_DEVICE = "/dev/qtec/video0"
""": Default video device for qtec cameras"""


class CameraEvent:
    """Frame available camera event

    An Event-like class that signals all active clients
    when a new frame is available.
    """

    CLIENT_TIMEOUT = 2
    WAIT_TIMEOUT = 2
    """: Timeout for considering the client as disconnected"""

    log: logging.Logger = logger_conf(__name__, c_level=logging.DEBUG)

    def __init__(self):
        """Constructor

        Initializes an empty list of events
        """
        self.events = {}

    def wait(self, timeout=WAIT_TIMEOUT):
        """Invoked from each client's thread to wait for the next frame."""
        ident = get_ident()
        if ident not in self.events:
            # this is a new client
            # add an entry for it in the self.events dict
            # each entry has two elements, a threading.Event() and a timestamp
            self.events[ident] = [threading.Event(), time.time()]
        return self.events[ident][0].wait(timeout)

    def cleanup(self):
        """Removes from event list all clients which timed out."""
        now = time.time()
        remove = []
        for ident, event in self.events.items():
            # if the client's event is already set, it means the client
            # did not process a previous frame
            # if the event stays set for more than X seconds, then assume
            # the client is gone and remove it
            if now - event[1] > CameraEvent.CLIENT_TIMEOUT:
                self.log.debug(f"Client timeout for {ident}")
                remove.append(ident)
        for ident in remove:
            del self.events[ident]

    def set(self):
        """Sets event (new frame available) for all client's threads.

        Invoked by the camera thread when a new frame is available.
        """
        now = time.time()
        remove = []
        for ident, event in self.events.items():
            if not event[0].is_set():
                # if this client's event is not set, then set it
                # also update the last set timestamp to now
                event[0].set()
                event[1] = now
            else:
                # if the client's event is already set, it means the client
                # did not process a previous frame
                # if the event stays set for more than X seconds, then assume
                # the client is gone and remove it
                if now - event[1] > CameraEvent.CLIENT_TIMEOUT:
                    self.log.debug(f"Client timeout for {ident}")
                    remove.append(ident)
        for ident in remove:
            del self.events[ident]

    def clear(self):
        """Clear event from client thread's list.

        Invoked from each client's thread after a frame was processed.
        """
        events = self.events.get(get_ident())
        if events and len(events) > 0:
            events[0].clear()

    def clear_list(self):
        """Remove all events (from all client threads) from list.

        Used when a streaming (camera) thread stops to reset events status.
        """
        self.log.debug(f"Clear list ({len(self.events)} items)")
        self.events.clear()


class Frame:
    """Encapsulates frame data and metadata

    :param data: frame bytes
    :param metadata: frame metadata
    :param timestamp: time of creation
    """

    JPEG_QUALITY = 85
    """: JPEG encoding quality setting"""

    def __init__(
        self,
        data: bytes,
        metadata: FrameMetadata,
        shape: Tuple[int, int],
        pix_fmt: str,
        timestamp: Optional[float] = None,
    ):
        """Constructor"""
        self.data = data
        self.metadata = metadata
        self.shape = shape
        if timestamp:
            self.time = timestamp
        else:
            self.time = time.time()
        self.pix_fmt = pix_fmt

    @staticmethod
    def jpeg(frame) -> "Frame":
        """
        Jpeg Encode frame

        :return: the JPEG encoded frame with metadata
        """
        if frame is None:
            return None
        data = frame.data

        match frame.pix_fmt:
            case "RGB3":
                pix_fmt = TJPF_RGB
            case "RGB6":
                # JPEG doesn't support 16-bit RGB, so present it with
                # something consumable
                data = (frame.data >> 8).astype(np.uint8)
                pix_fmt = TJPF_RGB
            case "BGR3":
                pix_fmt = TJPF_BGR
            case "RGB4":
                pix_fmt = TJPF_ARGB
            case "BGR4":
                pix_fmt = TJPF_BGRA
            case "GREY" | "QG08":
                pix_fmt = TJPF_GRAY
            case "Y16" | "QG16":
                data = (np.right_shift(frame.data, 8)).astype("uint8")
                pix_fmt = TJPF_GRAY
            case "GBRG" | "GRBG" | "RGGB" | "BA81":
                pix_fmt = TJPF_BGR
                match frame.pix_fmt:
                    case "GBRG":
                        data = cv2.cvtColor(frame.data, cv2.COLOR_BayerGR2BGR)
                    case "GRBG":
                        data = cv2.cvtColor(frame.data, cv2.COLOR_BayerGB2BGR)
                    case "RGGB":
                        data = cv2.cvtColor(frame.data, cv2.COLOR_BayerBG2BGR)
                    case "BA81":
                        data = cv2.cvtColor(frame.data, cv2.COLOR_BayerRG2BGR)
            case _:
                raise ValueError(f"Unsupported image format {frame.pix_fmt!r}")

        jpeg = TurboJPEG().encode(
            data, quality=Frame.JPEG_QUALITY, pixel_format=pix_fmt
        )
        return Frame(jpeg, frame.metadata, frame.shape, frame.pix_fmt, time.time())


class BaseCameraInfo:
    """Static variables for each video device"""

    def __init__(self):
        """Constructor"""
        self.thread = None  # background thread that reads frames from camera
        self.frame = None  # current frame is stored here by background thread
        self.frame_jpeg = None
        self.exception = None  # propagate information about last exception
        self.last_access = 0  # time of last client access to the camera
        self.event = CameraEvent()
        self.lock = threading.Lock()
        self.req_stop = False  # force stop


class BaseCamera:
    """Handles distributing images

    :param device: target video device
    :param stream: start streaming thread
    """

    devices: Dict[str, BaseCameraInfo] = dict()
    """: list of video devices"""
    STREAM_TIMEOUT: int = 3
    """: timeout to stop streaming when there are no incoming requests"""
    MAX_JPEG_FPS: int = 20
    """: throttle MJPEG to save CPU"""

    log: logging.Logger = logger_conf(__name__, c_level=logging.DEBUG)
    """: the class specific logger"""

    @staticmethod
    def add_device(device: str = DEF_DEVICE) -> None:
        """Add video device to internal list"""
        dev_path = get_video_device(device)
        if dev_path not in BaseCamera.devices:
            BaseCamera.log.debug("Adding dev=" + dev_path)
            BaseCamera.devices[dev_path] = BaseCameraInfo()

    @staticmethod
    def get_device(device: str = DEF_DEVICE) -> BaseCameraInfo:
        """Get video device data from internal list

        Adds new device to list if not present
        """
        dev_path = get_video_device(device)
        BaseCamera.add_device(dev_path)
        return BaseCamera.devices.get(dev_path)

    def __init__(self, device: str = DEF_DEVICE, stream: bool = False) -> None:
        """Constructor"""
        BaseCamera.add_device(device)
        self.video_device = device
        if stream:
            self.start_streaming(device)

    def start_streaming(self, device: str = DEF_DEVICE) -> None:
        """Start the background camera thread if it isn't running yet."""
        self.log.debug("start_streaming dev=" + device)
        dev = BaseCamera.get_device(device)
        if dev.thread is None:
            dev.last_access = time.time()
            dev.req_stop = False

            # start background frame thread
            dev.thread = threading.Thread(
                target=self._thread,
                args=(device,),
                name=f"{get_video_device(device)}_thread",
            )
            self.log.debug(f"Created thread: {dev.thread.name}")
            dev.thread.start()

            # wait until first frame is available
            # dev.event.wait()

    def get_frame(self) -> Frame:
        """Return the current camera frame."""
        dev = BaseCamera.get_device(self.video_device)
        dev.last_access = time.time()

        # wait for a signal from the camera thread
        if not dev.event.wait():
            # self.log.warning(f"Timeout waiting for frame from {self.video_device}")
            if dev.exception is not None:
                raise dev.exception
            else:
                raise V4L2Exception(
                    f"Timeout waiting for frame from {self.video_device}"
                )

        dev.event.clear()

        if dev.frame is None:
            raise V4L2Exception(f"No frame available from {self.video_device}")

        return dev.frame

    def get_frame_jpeg(self) -> Frame:
        """Return the current camera frame JPEG encoded.

        Note that this works fine for the sync mjpeg
        since each client has it's own thread and therefore it's own queue.
        However when async is used, several clients will share the same thread
        and therefore the same queue, meaning that they won't be able to get
        the same frame.
        """
        dev = BaseCamera.get_device(self.video_device)
        dev.last_access = time.time()

        # wait for a signal from the camera thread
        if not dev.event.wait():
            # self.log.warning(f"Timeout waiting for frame from {self.video_device}")
            if dev.exception is not None:
                raise dev.exception
            else:
                raise V4L2Exception(
                    f"Timeout waiting for frame from {self.video_device}"
                )

        dev.event.clear()

        if dev.frame is None:
            raise V4L2Exception(f"No frame available from {self.video_device}")

        # doing the jpeg conversion here instead of inside the thread
        # saves cpu since we don't need to encode frames unless requested
        # however it reduces the throughput to the client since it delays
        # the frame delivery by the time it takes to encode

        if dev.frame_jpeg is None:
            dev.frame_jpeg = Frame.jpeg(dev.frame)

        # update jpeg frame
        if dev.frame_jpeg.metadata.sequence != dev.frame.metadata.sequence:
            # throttle fps
            if time.time() - dev.frame_jpeg.time > 1 / BaseCamera.MAX_JPEG_FPS:
                dev.frame_jpeg = Frame.jpeg(dev.frame)

        return dev.frame_jpeg

    @staticmethod
    def frames(
        device: str = DEF_DEVICE,
    ) -> Generator[Tuple[Frame, Exception], None, None]:
        """Generator that returns frames from the camera."""
        raise RuntimeError("Must be implemented by subclasses.")

    @staticmethod
    def stop(device: str = DEF_DEVICE) -> None:
        """Stop streaming"""
        BaseCamera.log.debug("Requesting camera stop dev=" + device)
        dev = BaseCamera.get_device(device)
        dev.req_stop = True

    @staticmethod
    def stop_all() -> None:
        """Stop streaming from all video devices"""
        for device in BaseCamera.devices:
            BaseCamera.stop(device)

    @staticmethod
    def join_all_threads() -> None:
        """Join streaming threads from all video devices"""
        for device in BaseCamera.devices:
            dev = BaseCamera.get_device(device)
            if dev.thread is not None and dev.thread.is_alive():
                BaseCamera.log.debug(f"Joining thread {dev.thread.name}")
                dev.thread.join()

    @classmethod
    def _thread(cls, device: str = DEF_DEVICE) -> None:
        """Camera background streaming thread.

        One thread per video device
        """
        try:
            dev = BaseCamera.get_device(device)
            dev_str = get_video_device_str(device)
            # pix_fmt = None
            cls.log.info(f"Starting camera thread for {dev_str}")
            frames_iterator = cls.frames(device)
            for frame, exception in frames_iterator:
                if exception:
                    dev.exception = exception
                    cls.log.error(
                        f"Got exception from frame generator in {dev_str} thread: {exception}"
                    )
                    break
                else:
                    dev.exception = None

                # if pix_fmt is None:
                #     pix_fmt = cls.get_pixel_format().pixelformat

                dev.frame = frame
                dev.event.set()  # send signal to clients
                time.sleep(0)

                # if there hasn't been any clients asking for frames in
                # the last X seconds then stop the thread
                if time.time() - dev.last_access > BaseCamera.STREAM_TIMEOUT:
                    cls.log.info(
                        f"Stopping camera thread for {dev_str} due to inactivity."
                    )
                    break

                if dev.req_stop:
                    cls.log.info(f"Force Stopping camera thread for {dev_str}")
                    break

            frames_iterator.close()

        except Exception as e:
            cls.log.error(f"Unhandled exception in {dev_str} thread: {str(e)}")
            dev.exception = e

        finally:
            # clear pending events/frames
            dev.event.clear_list()
            dev.frame = None
            dev.frame_jpeg = None

            dev.req_stop = False
            dev.thread = None
