"""Quark API Backend for Qtec cameras

:Authors:
    Marianna Smidth Buschle <msb@qtec.com>
"""

import logging
import os.path
import sys
import threading
from typing import List
from typing import Optional

import uvicorn
from fastapi import FastAPI
from fastapi import Request
from fastapi.responses import HTMLResponse
from fastapi.responses import JSONResponse

from quark_backend.api.common import APIError
from quark_backend.api.common import app_logger
from quark_backend.api.controls import router as hl_controls_router
from quark_backend.api.device import router as device_router
from quark_backend.api.image import router as image_router
from quark_backend.api.oculus import OculusManager
from quark_backend.api.oculus import router as oculus_router
from quark_backend.api.stream import router as stream_router
from quark_backend.api.v4l2.control import router as control_router
from quark_backend.api.v4l2.controls import router as controls_router
from quark_backend.api.v4l2.crop import router as crop_router
from quark_backend.api.v4l2.framerate import router as framerate_router
from quark_backend.api.v4l2.pixel_format import router as pixel_format_router
from quark_backend.api.v4l2.resolution import router as resolution_router
from quark_backend.quark_camera import DEF_DEVICE
from quark_backend.quark_camera import Camera
from quark_backend.tcp_socket import TCPSocket
from quark_backend.utils import logger_conf


description = """
## Info on RAW streaming
The API provides RAW streaming through a TCP socket on port `12345`. When a new
client connects, the server sends a header describing the image format,
followed by a stream of frames.

The stream header is a binary header in big-endian format, comprised of the
following elements (in order):
- Header marker 'STREAM' literal (6 bytes)
- Frame height (2 bytes, uint16)
- Frame width (2 bytes, uint16)
- Frame depth (2 bytes, uint16) (1 if grayscale, 3 if RGB)
- Bytes per element (1 byte, uint8) (1 if 8-bit, 2 if 16-bit)
- End marker 'END' (3 bytes)

Each frame is represented as a byte array in row-major order.

### Frame message
Each frame starts with a header marker

- Marker 'F' (1 byte)
- Total header length (2 bytes, uint16)
- Sequence number (4 bytes, uint16)
- End marker 'E' (1 bytes)

Followed by the raw frame data

## HSI frame info
HSI frames are output in BIP format, i.e., each frame is a N_WIDTH x N_BANDS
sized byte array in row-major order.
"""

tags_meta = [
    {
        "name": "Controls",
        "description": "High level controls interface.",
    },
    {
        "name": "Frame Parameters",
        "description": "Video for linux frame settings.",
        "externalDocs": {
            "description": "Official docs",
            "url": "https://docs.qtec.com/camera-overview/configuration",
        },
    },
    {
        "name": "V4L2 Controls",
        "description": "Video for linux controls interface.",
        "externalDocs": {
            "description": "Official docs",
            "url": "https://docs.qtec.com/camera-overview/camera_controls",
        },
    },
    {
        "name": "Devices",
        "description": "List devices",
    },
    {
        "name": "Streaming",
        "description": "MJPEG streaming.",
    },
    {
        "name": "Images",
        "description": "Grab individual frames.",
    },
    {
        "name": "Oculus",
        "description": "Controls specific to the Oculus HSI camera",
    },
]

app = FastAPI(
    title="Quark API",
    description=description,
    summary="API for interacting with Qtechnology A/S cameras.",
    version="0.1.0",
    license_info={
        "name": "LGPL2.1",
        "url": "https://opensource.org/licenses/LGPL-2.1",
    },
    openapi_tags=tags_meta,
)

app.openapi_version = "3.0.3"

# include routers
app.include_router(device_router)
app.include_router(control_router)
app.include_router(controls_router)
app.include_router(framerate_router)
app.include_router(pixel_format_router)
app.include_router(resolution_router)
app.include_router(crop_router)
app.include_router(stream_router)
app.include_router(image_router)
app.include_router(hl_controls_router)

IS_OCULUS = os.path.isfile(os.path.join(os.path.expanduser("~"), "oculus_config.toml"))

if IS_OCULUS:
    app_logger.info("Oculus settings enabled")
    app.include_router(oculus_router)
else:
    app_logger.info("Oculus settings disabled")


@app.exception_handler(APIError)
async def unicorn_exception_handler(request: Request, exc: APIError):
    """Return custom JSON when APIError or its children are raised"""
    response = {"error": exc.description, "message": ""}
    if len(exc.args) > 0:
        response["message"] = exc.args[0]
    app_logger.error(f"{exc.description}: {response['message']}")
    return JSONResponse(
        status_code=exc.code,
        content={"message": response},
    )


@app.exception_handler(500)
async def unicorn_exception_handler(request: Request, exc: APIError):  # noqa: F811
    """Return JSON instead of HTML for any other server error"""
    app_logger.error(f"Unknown Exception: {str(exc)}")
    # app_logger.debug(
    #    "".join(traceback.format_exception(err, value=err, tb=err.__traceback__))
    # )
    response = {"error": "Unhandled Internal Server Error"}
    return JSONResponse(
        status_code=500,
        content={"message": response},
    )


@app.on_event("shutdown")
def shutdown_event():
    """Shutdown event"""
    app_logger.debug("Shutdown event")


def handle_args(cli_args: Optional[List[str]] = None) -> List[str]:
    """Command-line arguments

    `cli_args` makes it possible to call functions command-line-style
    from other Python code without touching sys.argv.
    """
    if cli_args is None:
        return sys.argv[1:]
    else:
        return cli_args


def generate_index_html():
    html_content = """
    <html>
        <head>
            <title>Quark Backend API</title>
        </head>
        <body>
            <h1>API for interacting with Qtechnology A/S cameras.</h1>
            <ul>
                <li><a href="docs">Live API documentation</a></li>
                <li><a href="mjpeg">MJPEG stream</a></li>
                <li><a href="http://docs.qtec.com">Official qtec docs</a></li>
            </ul>
        </body>
    </html>
    """
    return HTMLResponse(content=html_content, status_code=200)


@app.get("/", response_class=HTMLResponse)
def home() -> HTMLResponse:
    """index.html"""
    return generate_index_html()


def main(cli_args: Optional[List[str]] = None):
    """Fast API App

    `cli_args` makes it possible to call this function command-line-style
    from other Python code without touching sys.argv.
    """
    log = logger_conf("main", c_level=logging.DEBUG)

    # app.logger.warning("Test")

    # start socket thread
    log.debug("Starting socket thread...")
    socket = TCPSocket(DEF_DEVICE, mjpeg=False)
    t = threading.Thread(target=socket.socket_thread)
    t.start()

    if IS_OCULUS:
        log.info("Applying Oculus settings")
        OculusManager.get_config().apply_settings()
        OculusManager.set_socket(socket)

    DEBUG_MODE = False

    # start server thread
    log.debug("Starting Uvicorn server thread")
    uvicorn.run(
        "quark_backend.main:app",
        host="0.0.0.0",  # noqa: S104
        port=5000,
        log_level="debug" if DEBUG_MODE else "info",
        reload=DEBUG_MODE,
        # reload_dirs=[],
    )

    log.debug(f"Current number of threads: {threading.active_count()}")
    Camera.stop_all()
    Camera.join_all_threads()
    socket.stop()
    t.join()
    socket.join_all_threads()


if __name__ == "__main__":
    main()
