import os.path
import tomllib
from threading import Lock

import numpy as np
import qamlib
import toml
from pydantic import BaseModel
from pydantic import Field
from pydantic import RootModel
from quarklib import Camera as QCamera


try:
    from sklearn.cluster import AgglomerativeClustering

    HAS_SKLEARN = True
except ImportError:
    HAS_SKLEARN = False

from quark_backend.quark_camera import Camera
from quark_backend.tcp_socket import TCPSocket


HOME_DIR = os.path.expanduser("~")
CONFIG_PATH = os.path.join(HOME_DIR, "oculus_config.toml")
SETTINGS_PATH = os.path.join(HOME_DIR, "oculus_settings.json")


class SpectrographConfig(BaseModel):
    """Spectrograph parameters that are universal to each Oculus camera type."""

    spectral_range_nm: tuple[int, int] = Field(example=(430, 1000))
    sensor_crop_height: int = Field(ge=0, example=350)
    sensor_crop_width: int = Field(ge=0, example=1840)

    def spectral_pixel_resolution(self):
        """Get the spectral resolution per pixel, given in nm/pixel."""
        return (
            self.spectral_range_nm[1] - self.spectral_range_nm[0]
        ) / self.sensor_crop_height


class SpectrographCalibration(BaseModel):
    """Calibration settings that are unique to each camera."""

    crop_top: int = Field(ge=0, example=1050)
    crop_left: int = Field(ge=0, example=108)


class SpectralBandSelection(RootModel):
    """Defines the subset of spectral bands to capture."""

    root: list[int | tuple[int, int] | tuple[int, int, int]] = Field(
        example=[10, [20, 30], [50, 200, 5]]
    )

    def line_indexes(self) -> list[int]:
        """Calculate full list of line indices for a band selection."""
        out = []
        for b in self.root:
            if isinstance(b, tuple):
                if len(b) == 2:
                    out.extend(list(range(b[0], b[1])))
                else:
                    out.extend(list(range(b[0], b[1], b[2])))
            else:
                out.append(b)
        out = sorted(out)
        return out


class OculusConfig(BaseModel):
    """Describes all configuration parameters for a specific Oculus camera."""

    spectrograph: SpectrographConfig
    calibration: SpectrographCalibration


class HorizontalCrop(BaseModel):
    """A horizontal crop defined by a start and end x-position."""

    start: int = Field(example=50)
    end: int = Field(example=1500)


class OculusSettings(BaseModel):
    """User defined Oculus camera settings."""

    bands: SpectralBandSelection
    horizontal_crop: HorizontalCrop | None = None
    horizontal_binning: int = 1
    vertical_binning: int = 1
    exposure: int | list[int] = Field(example=[800] * 350)


class OculusSettingsRequest(BaseModel):
    """Settings request. All fields are optional."""

    bands: SpectralBandSelection | None = Field(
        default=None, description="Spectral band selection"
    )
    horizontal_crop: HorizontalCrop | None = Field(
        default=None, description="Horizontal crop selection"
    )
    horizontal_binning: int | None = Field(
        default=None, description="Horizontal binning"
    )
    vertical_binning: int | None = Field(default=None, description="Vertical binning")
    exposure: int | list[int] | None = Field(
        default=None, ge=14, example=500, description="Exposure selection"
    )


def simplify_crops(crops: list[tuple[int, int]]) -> list[tuple[int, int]]:
    out = []
    i = 0
    while i < len(crops) - 1:
        a = crops[i]
        for j in range(i + 1, len(crops)):
            b = crops[j]
            if a[1] + 1 == b[0]:
                a = (a[0], b[1])
                i += 1
            else:
                break
        i += 1
        out.append(a)
    return out


def fix_crops_to_limits(crops: list[tuple[int, int]]) -> list[tuple[int, int]]:
    out = []
    for c in crops:
        out.append((c[0] // 4 * 4, (c[1] + 3) // 4 * 4))
    return out


def check_crop_coverage(crops: list[tuple[int, int]], rects: list[qamlib.Rectangle]):
    for c, r in zip(crops, rects):
        if r.top > c[0] or r.top + r.height < c[1]:
            raise ValueError(f"Tried to set crop: {c}, got {r}")


class OculusCamera(BaseModel):
    """The Oculus camera model describing both configuration parameters and user settings.

    The model applies user settings to the sensor based on the current calibration. The
    consequence is that settings are not camera-dependent but should give identical
    results on cameras with different calibrations.
    """

    config: OculusConfig
    settings: OculusSettings

    def _get_calibration_top(self) -> int:
        return int(self.config.calibration.crop_top / self.settings.vertical_binning)

    def _get_calibration_left(self) -> int:
        return int(self.config.calibration.crop_left / self.settings.horizontal_binning)

    def _get_sensor_height(self) -> int:
        return int(
            self.config.spectrograph.sensor_crop_height / self.settings.vertical_binning
        )

    def _get_sensor_width(self) -> int:
        return int(
            self.config.spectrograph.sensor_crop_width
            / self.settings.horizontal_binning
        )

    def _get_crop_start(self) -> int:
        start = (
            0
            if self.settings.horizontal_crop is None
            else self.settings.horizontal_crop.start
        )
        return int(start / self.settings.horizontal_binning)

    def _get_crop_end(self) -> int:
        end = (
            self.config.spectrograph.sensor_crop_width
            if self.settings.horizontal_crop is None
            else self.settings.horizontal_crop.end
        )
        return int(end / self.settings.horizontal_binning)

    @classmethod
    def default(cls):
        with open(CONFIG_PATH, "rb") as f:
            cfg = tomllib.load(f)
        return OculusCamera(
            config=OculusConfig.parse_obj(cfg),
            settings=OculusSettings(
                bands=[(0, 350)],
                exposure=1813,
            ),
        )

    @classmethod
    def load_from_file(cls):
        with open(CONFIG_PATH, "rb") as f:
            cfg = tomllib.load(f)
        return OculusCamera(
            config=OculusConfig.parse_obj(cfg),
            settings=OculusSettings.parse_file(SETTINGS_PATH),
        )

    def wavelengths(self) -> np.ndarray:
        lines = np.array(self.settings.bands.line_indexes())

        resolution = (
            self.config.spectrograph.spectral_pixel_resolution()
            * self.settings.vertical_binning
        )
        return (
            lines + self.settings.vertical_binning / 2
        ) * resolution + self.config.spectrograph.spectral_range_nm[0]

    def line_order(self) -> list[int]:
        lines = self.settings.bands.line_indexes()
        lines = [
            list(
                range(
                    i * self.settings.vertical_binning,
                    (i + 1) * self.settings.vertical_binning,
                )
            )
            for i in lines
        ]
        lines = np.concatenate(lines)
        print("ll", len(lines))
        return [line + self._get_calibration_top() for line in lines]

    def get_max_crop(self):
        width = (
            self.config.spectrograph.sensor_crop_width
            // self.settings.horizontal_binning
        )
        height = (
            self.config.spectrograph.sensor_crop_height
            // self.settings.vertical_binning
        )
        return width, height

    def apply_settings(self):
        cam = Camera._get_camera("/dev/qtec/video0")

        cam.set_crop(0, 0, 10, 10)

        cam.set_control("horizontal_binning", self.settings.horizontal_binning)
        cam.set_control("vertical_binning", self.settings.vertical_binning)
        cam_type = cam.get_ext_control("sensor_type")

        lines = self.settings.bands.line_indexes()

        if cam_type.value == "IMX990 Mono":
            cam.controls.capture.exposure_time.write(self.settings.exposure)

            if not HAS_SKLEARN:
                raise ImportError(
                    "Missing sklearn library. Install with `pip install scikit-learn`"
                )
            model = AgglomerativeClustering(n_clusters=8)  # What if not possible?
            model.fit(np.array(lines).reshape(-1, 1))
            labels = model.labels_
            unique, unique_pos = np.unique(labels, return_index=True)
            unique = unique[np.argsort(unique_pos)]

            band_select = []
            for label in unique:
                group = np.array(lines)[labels == label]
                min_val = group.min()
                max_val = group.max()
                band_select.append((min_val, max_val))

            band_select = simplify_crops(band_select)
            band_select = fix_crops_to_limits(band_select)
            print(band_select)

            # if len(self.settings.bands.__root__) > 8:
            #     raise ValueError(
            #         f"Sensor type {cam_type.value} can only handle a maximum of 8 simultaneous band regions."
            #     )

            rects = []
            crop_left = self._get_calibration_left() + max(self._get_crop_start(), 0)
            crop_right = self._get_calibration_left() + min(
                self._get_crop_end(), self._get_sensor_width()
            )
            total_height = 0
            for s in band_select:
                if not isinstance(s, tuple) or len(s) != 2:
                    raise ValueError(
                        f"Sensor type {cam_type.value} only supports continuous band ranges, got either individual or strided values."
                    )

                print(crop_left, s[0], crop_right - crop_left, s[1] - s[0])
                rects.append(
                    qamlib.Rectangle(
                        crop_left, s[0], crop_right - crop_left, s[1] - s[0]
                    )
                )
                total_height += s[1] - s[0]

            cam.set_resolution(crop_right - crop_left, total_height)
            cam.set_crop(rects)
            print("ddrr", rects)

            check_crop_coverage(band_select, cam.get_crop())

        else:
            crop_height = self._apply_crop(cam, lines)

            if len(lines) > self.config.spectrograph.sensor_crop_height:
                raise ValueError("Too many bands selected for the given crop height")
            self._apply_exposure_lines(cam, crop_height, lines)
            self._apply_band_selection(cam, lines)

    @staticmethod
    def _apply_band_selection(cam: QCamera, lines: list[int]):
        """Apply band selection for the GSENSE sensor."""
        lines = np.uint32(lines)
        cam.controls.exposure_sequence.line_order._input_shape = lines.shape
        cam.controls.exposure_sequence.line_order.write(lines)

    def _apply_crop(self, cam: QCamera, lines):
        """Calculate and apply sensor crop based on configuration and settings.

        The sensor crop depends on both the sensor type and line selection.
        Currently, only GSENSE is supported, but the method will be able to
        handle IMX990 as well in the future.

        For GSENSE, the crop height is the number of lines selected for the
        'bands' setting. The crop position and width is determined by the
        camera calibration and configuration.
        """
        crop_height = (
            min(len(lines), self.config.spectrograph.sensor_crop_height)
            // self.settings.vertical_binning
        )
        if self.settings.horizontal_crop is None:
            crop_left = self._get_calibration_left()
            crop_right = self._get_calibration_left() + self._get_sensor_width()
        else:
            crop_left = self._get_calibration_left() + max(self._get_crop_start(), 0)
            crop_right = self._get_calibration_left() + min(
                self._get_crop_end(), self._get_sensor_width()
            )
            print(crop_right)

        print(crop_left, crop_right, crop_height)
        cam.set_crop(
            crop_left,
            self._get_calibration_top(),
            crop_right - crop_left,
            crop_height,
        )
        print(crop_height, crop_left, self._get_calibration_top(), cam.get_crop())
        actual_crop = cam.get_crop()[0]
        new_start = (
            actual_crop.left * self.settings.horizontal_binning
            - self.config.calibration.crop_left
        )
        new_end = new_start + actual_crop.width * self.settings.horizontal_binning
        print(new_start, new_end)
        self.settings.horizontal_crop = HorizontalCrop(start=new_start, end=new_end)
        return crop_height

    def _apply_exposure_lines(self, cam: QCamera, crop_height: int, lines: list[int]):
        """Apply exposure settings.

        If a global value is used, the exposure is set to that value for each line."""
        if isinstance(self.settings.exposure, int):
            exp = np.ones((crop_height, 1), dtype=np.uint32) * self.settings.exposure
        else:
            exp = np.array(self.settings.exposure, dtype=np.uint32).reshape((350, 1))
            if len(exp) != crop_height:
                raise ValueError("Exposure sequence invalid")
        cam.controls.exposure_sequence.line_exposure_sequence[
            lines[0] : lines[0] + crop_height
        ] = exp


class OculusManager:
    """Oculus camera manager.

    Is responsible for loading/saving the settings and configuration files."""

    _config: OculusCamera | None = None
    _tcp: TCPSocket | None = None
    lock: Lock = Lock()

    @classmethod
    def get_config(cls) -> OculusCamera:
        if cls._config is None:
            if os.path.isfile(SETTINGS_PATH):
                cls._config = OculusCamera.load_from_file()
            else:
                cls._config = OculusCamera.default()
                cls.save_settings(cls._config.settings)

        return cls._config

    @classmethod
    def set_socket(cls, socket: TCPSocket):
        cls._tcp = socket

    @classmethod
    def set_preprocessing(cls, func):
        cls._tcp.post_process = func

    @classmethod
    def save_config(cls, value: OculusConfig):
        cls._config.config = value
        with open(CONFIG_PATH, "w") as file:
            toml.dump(value.dict(), file)

    @classmethod
    def save_settings(cls, value: OculusSettings):
        cls._config.settings = value
        with open(SETTINGS_PATH, "w") as file:
            file.write(value.json())
