"""Quark Camera"""

import json
import logging
import time
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy
import qamlib
from pydantic import BaseModel
from pydantic import Field
from quarklib import Camera as QCamera

from quark_backend.base_camera import DEF_DEVICE
from quark_backend.base_camera import BaseCamera
from quark_backend.base_camera import Frame
from quark_backend.utils import get_video_device
from quark_backend.utils import logger_conf


class MyJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder"""

    def default(self, obj):
        """Calls to_json() method of object"""
        return obj.to_json()


class FrameSize(BaseModel):
    """Frame size model"""

    width: int = Field(ge=0, example=1024)
    height: int = Field(ge=0, example=1024)


class FrameResolution(BaseModel):
    """Video device frame resolution model"""

    width: int | None = Field(ge=0, example=256)
    height: int | None = Field(ge=0, example=256)


class CropRect(FrameSize):
    """Video device frame crop rectangle model"""

    left: int = Field(ge=0, example=100)
    top: int = Field(ge=0, example=100)

    class Config:
        schema_extra = {
            "examples": [
                {
                    "left": 100,
                    "top": 100,
                    "width": 100,
                    "height": 100,
                },
                {
                    "left": 0,
                    "top": 0,
                    "width": 500,
                    "height": 600,
                },
            ]
        }


class ImageFormatFlags(BaseModel):
    """Image Format flags model"""

    compressed: bool = Field(default=False)
    continuous_bytestream: bool = Field(default=False)
    csc_colorspace: bool = Field(default=False)
    csc_hsv_enc: bool = Field(default=False)
    csc_quantization: bool = Field(default=False)
    csc_xfer_func: bool = Field(default=False)
    csc_ycbcr_enc: bool = Field(default=False)
    dyn_resolution: bool = Field(default=False)
    emulated: bool = Field(default=False)
    enc_cap_frame_interval: bool = Field(default=False)


class PixelFormat(BaseModel):
    """Pixel Format model"""

    pixelformat: str = Field(
        default=...,
        description=(
            "FourCC representation of pixel format. "
            "Append '_BE' for big endian formats."
        ),
        example="RGB3",
        max_length=7,
        min_length=3,
    )
    """: FourCC representation of pixel format. Append '_BE' for big endian formats"""


class PixelFormatFlags(BaseModel):
    """Pixel Format flags model"""

    premul_alpha: bool = Field(default=False)
    set_csc: bool = Field(default=False)


class SinglePlaneFmt(PixelFormat, FrameSize):
    """Single Plane Format model"""

    size: int = Field(default=..., example=307200)
    bytes_per_line: int = Field(default=..., example=300)
    priv: bool = Field(default=..., example=True)
    flags: Optional[PixelFormatFlags] = Field(
        default=None, example={"premul_alpha": False, "set_csc": False}
    )
    field: Optional[str] = Field(default=None, example=None)
    colorspace: str = Field(default="SRGB", example="SRGB")

    class Config:
        """Config class"""

        extra = "allow"
        """: will automatically add any fields not present into the model"""


class ImageFormat(PixelFormat):
    """Image Format model"""

    description: str = Field(default=..., example="24-bit RGB 8-8-8")
    """: Description of pixel format"""
    index: Optional[int] = Field(default=None)
    # flags: Optional[ImageFormatFlags] = Field(default=None)


class FrameRate(BaseModel):
    """Framerate value model"""

    value: float = Field(default=..., example=20)


class DiscreteFrameRate(BaseModel):
    """Discrete Framerate model"""

    values: List[float] = Field(
        default=...,
        description=(
            "Discrete list of possible framerates."
            "Dependant on the current frame size and format"
        ),
        example=[10, 20, 30],
    )


class ContinuousFrameRate(BaseModel):
    """Continuous Framerate model"""

    min: float = Field(default=..., description="minimum framerate allowed", example=1)
    """: Minimum framerate"""
    max: float = Field(default=..., description="maximum framerate allowed", example=30)
    """: Maximum framerate"""


class StepwiseFrameRate(ContinuousFrameRate):
    """Stepwise Framerate model"""

    step: float = Field(default=..., description="framerate step", example=1)
    """: Framerate step"""


class TriggerSequence(BaseModel):
    """Single Trigger Sequence model"""

    exposure_time: int
    flash_time: int
    frame_delay: int
    trigger_delay: int


class TriggerSequenceValue(BaseModel):
    """Trigger Sequence value model"""

    trig_seq_list: List[TriggerSequence]
    """: list of trigger sequences"""


class V4L2ControlValue(BaseModel):
    """V4l2 control value model"""

    value: Union[int, str, List[int], TriggerSequenceValue, Any]
    """: value of v4l2 control"""

    class Config:
        """Config class"""

        schema_extra = {
            "examples": [
                {
                    "value": 5000,
                },
            ]
        }


def new_control_value_obj(model_obj: V4L2ControlValue) -> qamlib.ControlValue:
    """
    Converts pydantic model to qamlib.ControlValue subclasses

    Necessary for qamlib.Camera.set_ext_control() and qamlib.Camera.set_controls()

    :param model_obj: Pydantic BaseModel ControlValue
    :return: object which is a subclass of qamlib.ControlValue
    """
    if isinstance(model_obj.value, int):
        return qamlib.IntegerControlValue(model_obj.value)
    elif isinstance(model_obj.value, str):
        return qamlib.StringControlValue(model_obj.value)
    elif isinstance(model_obj.value, list):
        return qamlib.ArrayControlValue(numpy.array(model_obj.value, dtype=numpy.int32))
    elif isinstance(model_obj.value, TriggerSequenceValue):
        ctrl_val = qamlib.TriggerSequenceValue()
        for s in model_obj.value.trig_seq_list:
            ctrl_val.add_sequence(
                s.exposure_time,
                s.flash_time,
                s.frame_delay,
                s.trigger_delay,
            )
        return ctrl_val
    else:
        raise TypeError(f"Unsupported type: {type(model_obj.value)}")


class V4L2ControlFlags(BaseModel):
    """V4l2 control flags model"""

    disabled: bool = Field(default=False)
    grabbed: bool = Field(default=False)
    read_only: bool = Field(default=False)
    update: bool = Field(default=False)
    inactive: bool = Field(default=False)
    slider: bool = Field(
        default=False,
        example=True,
        description="Control can be represented as a slider",
    )
    write_only: bool = Field(default=False)
    is_volatile: bool = Field(default=False)
    has_payload: bool = Field(default=False)
    execute_on_write: bool = Field(default=False)
    modify_layout: bool = Field(default=False)


class V4L2Control(BaseModel):
    """V4l2 control model"""

    id: int = Field(example=9963790)
    name: str = Field(example="Red Balance")
    type: str = Field(example="Int")
    min: Optional[int] = Field(default=None, example=0)
    max: Optional[int] = Field(default=None, example=65535)
    step: Optional[int] = Field(default=None, example=1)
    default_value: Optional[int] = Field(default=None, example=16384)
    items: Optional[Dict] = Field(
        default=None,
        description=(
            "mapping of items of menu controls."
            'In the form: "<menu_item_value>": "<menu_item_name>"'
        ),
        example={"0": "menu_value0", "1": "menu_value1", "3": "menu_value3"},
    )
    flags: Optional[V4L2ControlFlags] = Field(default=None)
    elements: Optional[int] = Field(
        default=None,
        description="number of elements in array controls",
        example=2048,
    )
    element_size: Optional[int] = Field(
        default=None,
        description=("size (in bytes) of each element in the array of array controls"),
        example=4,
    )
    dimensions: Optional[List[int]] = Field(
        default=None,
        description=("array shape in array controls"),
        example=[2048],
    )


def _measure_fps(n_frames: int, old_time: float) -> Tuple[int, float]:
    """Measure the frame 'readout' FPS"""
    n_frames = n_frames + 1
    time_now = time.time()
    if time_now - old_time > 1:
        time_now = time.time()
        BaseCamera.log.info("FPS: " + str(n_frames / (time_now - old_time)))
        old_time = time_now
        n_frames = 0
    return n_frames, old_time


class Camera(BaseCamera):
    """Qamlib based camera

    :param device: target video device
    :param stream: start streaming thread
    """

    quark_cameras: Dict[str, QCamera] = dict()
    """: list of qamlib cameras"""
    log: logging.Logger = logger_conf(__name__, c_level=logging.DEBUG)
    """: the class specific logger"""

    def __init__(self, device: str = DEF_DEVICE, stream: bool = False) -> None:
        """Constructor"""
        Camera._add_camera(device)
        super().__init__(device, stream)

    @staticmethod
    def _add_camera(device: str = DEF_DEVICE) -> None:
        """Add camera to internal list"""
        device = get_video_device(device)
        if device not in Camera.quark_cameras:
            Camera.quark_cameras[device] = QCamera(device)

    @staticmethod
    def _get_camera(device: str = DEF_DEVICE) -> QCamera:
        """Get camera from internal list"""
        device = get_video_device(device)
        Camera._add_camera(device)
        return Camera.quark_cameras.get(device)

    @staticmethod
    def frames(
        device: str = DEF_DEVICE,
    ) -> Generator[Tuple[Frame | None, Exception], None, None]:
        """Generator that returns frames from the camera."""
        try:
            Camera.log.debug("Starting frame capture dev=" + device)
            cam = Camera._get_camera(device)
            cam.start()
            pix_fmt = cam.get_format().pixelformat.fourcc
            fps = cam.get_framerate()
            timeout = int(1 / fps) + 1
            n_frames = 0
            old_time = time.time()
            retry = False
            retries = 5
            while True:
                try:
                    if retry:
                        if retries <= 0:
                            Camera.log.error("No more retries")
                            break

                        Camera.log.info(
                            f"Retrying streaming in 1 second: {retries} left"
                        )
                        retries -= 1
                        time.sleep(1)
                        cam.start()
                        # update the pixel format if the streaming was restarted
                        Camera.log.info(f"new format {cam.get_format().pixelformat}")
                        pix_fmt = cam.get_format().pixelformat.fourcc

                    # read current frame
                    metadata, frame = cam.get_frame(timeout)
                    retry = False
                    retries = 5

                    n_frames, old_time = _measure_fps(n_frames, old_time)

                    yield Frame(frame, metadata, frame.shape[:2], pix_fmt), None
                except (RuntimeError, qamlib.TimeoutException) as e:
                    Camera.log.debug(f"Caught exception: {e}")
                    retry = True

                # Camera.log.debug(metadata)
                # Camera.log.debug(metadata.sequence)
                # Camera.log.debug(frame.shape)

        except Exception as e:
            # Camera.log.debug(f"Caught exception in frame generator: {e}")
            Camera.log.exception(e)
            yield None, e
        finally:
            cam.stop()

    @staticmethod
    def list_controls(device: str = DEF_DEVICE) -> List[V4L2Control]:
        """List v4l2 controls from video device"""
        try:
            cam = Camera._get_camera(device)
            ctrls = cam.list_controls()
        except Exception as e:
            raise e

        # return [V4L2Control(**ctrls[key].to_json()) for key in ctrls]
        return [V4L2Control.parse_obj(ctrls[key].to_json()) for key in ctrls]
        # return json.dumps(ctrls, cls=MyJSONEncoder)

    @staticmethod
    def get_control(
        ctrl_id_or_name: Union[int, str], device: str = DEF_DEVICE
    ) -> V4L2ControlValue:
        """Get v4l2 control value from video device"""
        try:
            cam = Camera._get_camera(device)
            ctrl_val = cam.get_ext_control(ctrl_id_or_name)
        except Exception as e:
            raise e

        return V4L2ControlValue(**ctrl_val.to_json())

    @staticmethod
    def set_control(
        ctrl_id_or_name: Union[int, str],
        ctrl_value: V4L2ControlValue,
        device: str = DEF_DEVICE,
    ) -> V4L2ControlValue:
        """Set v4l2 control value on video device"""
        try:
            cam = Camera._get_camera(device)
            cam.set_ext_control(ctrl_id_or_name, new_control_value_obj(ctrl_value))
            ctrl_val = cam.get_ext_control(ctrl_id_or_name)
        except Exception as e:
            raise e

        return V4L2ControlValue(**ctrl_val.to_json())

    @staticmethod
    def get_controls(
        ctrl_list: Optional[List[str]], device: str = DEF_DEVICE
    ) -> Dict[str, V4L2ControlValue]:
        """Get v4l2 control value from video device"""
        try:
            cam = Camera._get_camera(device)
            if ctrl_list is None:
                ctrls = cam.list_controls()
                ctrl_list = []
                for key in ctrls:
                    if ctrls[key].type == qamlib.ControlType.CTRL_CLASS:
                        continue
                    if ctrls[key].flags.write_only:
                        continue
                    if ctrls[key].flags.execute_on_write:
                        continue
                    ctrl_list.append(ctrls[key].name)
            value_list = cam.get_controls(ctrl_list)
        except Exception as e:
            raise e

        return {key: value_list[key].to_json() for key in value_list}

    @staticmethod
    def set_controls(
        ctrl_list: Dict[str, V4L2ControlValue], device: str = DEF_DEVICE
    ) -> Dict[str, V4L2ControlValue]:
        """Get v4l2 control value from video device"""
        try:
            cam = Camera._get_camera(device)

            qamlib_list = dict()
            for key in ctrl_list:
                qamlib_list[key] = new_control_value_obj(ctrl_list[key])

            cam.set_controls(qamlib_list)
            value_list = cam.get_controls([key for key in ctrl_list])
        except Exception as e:
            raise e

        return {key: value_list[key].to_json() for key in value_list}

    @staticmethod
    def get_resolution(device: str = DEF_DEVICE) -> FrameResolution:
        """Get frame resolution from video device"""
        try:
            cam = Camera._get_camera(device)
            width, height = cam.get_resolution()
        except Exception as e:
            raise e

        return FrameResolution(width=width, height=height)

    @staticmethod
    def set_resolution(
        width: int = None, height: int = None, device: str = DEF_DEVICE
    ) -> FrameResolution:
        """Set frame resolution on video device"""
        try:
            cam = Camera._get_camera(device)
            if height is None or width is None:
                w, h = cam.get_resolution()
                if height is None:
                    height = h
                if width is None:
                    width = w

            # auto-crop to center
            bounds = cam.get_crop_bounds()
            left = int(((bounds.width - bounds.left) - width) / 2) + bounds.left
            top = int(((bounds.height - bounds.top) - height) / 2) + bounds.top

            Camera.log.debug(
                f"Set_resolution dev={device}: {width}x{height} left:{left} top:{top}"
            )
            cam.set_crop(left, top, width, height)

            rects = cam.get_crop()
        except Exception as e:
            raise e

        return [
            CropRect(width=r.width, height=r.height, top=r.top, left=r.left)
            for r in rects
        ]

    @staticmethod
    def get_crop_bounds(device: str = DEF_DEVICE) -> CropRect:
        """Get frame crop bounds from video device"""
        try:
            cam = Camera._get_camera(device)
            r = cam.get_crop_bounds()
        except Exception as e:
            raise e

        return CropRect(width=r.width, height=r.height, top=r.top, left=r.left)

    @staticmethod
    def get_crop_def(device: str = DEF_DEVICE) -> CropRect:
        """Get frame crop bounds from video device"""
        try:
            cam = Camera._get_camera(device)
            r = cam.get_crop_default()
        except Exception as e:
            raise e

        return CropRect(width=r.width, height=r.height, top=r.top, left=r.left)

    @staticmethod
    def get_crop(device: str = DEF_DEVICE) -> List[CropRect]:
        """Get frame crop from video device"""
        try:
            cam = Camera._get_camera(device)
            rects = cam.get_crop()
        except Exception as e:
            raise e

        return [
            CropRect(width=r.width, height=r.height, top=r.top, left=r.left)
            for r in rects
        ]

    @staticmethod
    def set_crop(rects: List[CropRect], device: str = DEF_DEVICE) -> List[CropRect]:
        """Set frame crop on video device"""
        try:
            cam = Camera._get_camera(device)
            cam.set_crop(
                [
                    qamlib.Rectangle(
                        r.left,
                        r.top,
                        r.width,
                        r.height,
                    )
                    for r in rects
                ]
            )
            rects = cam.get_crop()
        except Exception as e:
            raise e

        return [
            CropRect(width=r.width, height=r.height, top=r.top, left=r.left)
            for r in rects
        ]

    @staticmethod
    def get_pixel_format(device: str = DEF_DEVICE) -> SinglePlaneFmt:
        """Get pixel format from video device"""
        try:
            cam = Camera._get_camera(device)
            fmt = cam.get_format()
        except Exception as e:
            raise e

        return SinglePlaneFmt(**fmt.to_json())

    @staticmethod
    def list_pixel_formats(device: str = DEF_DEVICE) -> List[ImageFormat]:
        """List pixel formats on video device"""
        try:
            cam = Camera._get_camera(device)
            fmts = cam.list_formats()
        except Exception as e:
            raise e

        return [ImageFormat(**fmts[key].to_json()) for key in fmts]

    @staticmethod
    def set_pixel_format(pixel_format, device: str = DEF_DEVICE) -> SinglePlaneFmt:
        """Set pixel format on video device"""
        try:
            cam = Camera._get_camera(device)
            cam.set_format(pixel_format)
            fmt = cam.get_format()
        except Exception as e:
            raise e

        return SinglePlaneFmt(**fmt.to_json())

    @staticmethod
    def get_framerate(device: str = DEF_DEVICE) -> FrameRate:
        """Get framerate from video device"""
        try:
            cam = Camera._get_camera(device)
            fps = cam.get_framerate()
        except Exception as e:
            raise e

        return FrameRate(value=fps)

    @staticmethod
    def list_framerates(
        device: str = DEF_DEVICE,
    ) -> Union[DiscreteFrameRate, ContinuousFrameRate, StepwiseFrameRate]:
        """
        List framerate limits on video device

        Depends on the current pixel format and frame size
        """
        try:
            cam = Camera._get_camera(device)
            fps_limits = cam.get_framerates()
        except Exception as e:
            raise e

        if isinstance(fps_limits, qamlib.StepwiseFrameRate):
            return StepwiseFrameRate(**fps_limits.to_json())
        elif isinstance(fps_limits, qamlib.ContinuousFrameRate):
            return ContinuousFrameRate(**fps_limits.to_json())
        elif isinstance(fps_limits, qamlib.DiscreteFrameRate):
            return DiscreteFrameRate(**fps_limits.to_json())
        else:
            raise TypeError(f"Unsupported type: {type(fps_limits)}")

    @staticmethod
    def set_framerate(framerate, device: str = DEF_DEVICE) -> FrameRate:
        """Set framerate on video device"""
        try:
            cam = Camera._get_camera(device)
            fps = cam.set_framerate(framerate)
        except Exception as e:
            raise e

        return FrameRate(value=fps)
