"""TCP socket"""

import logging
import socket
import threading

from qamlib import FrameMetadata

from quark_backend.base_camera import DEF_DEVICE
from quark_backend.quark_camera import Camera
from quark_backend.utils import logger_conf


# raw:
# vlc --demux rawvideo --rawvid-chroma RV24 --rawvid-fps 20 \
# --rawvid-width 1024 --rawvid-height 1024 tcp://10.10.150.225:12345

# jpeg:
# vlc tcp://10.10.150.225:12345


class TCPSocket:
    """TCPSocket class

    :param device: target video device
    :param mjeg: produce raw or mjpeg encoded frames
    :param port: TCP port
    """

    log: logging.Logger = logger_conf(__name__, c_level=logging.DEBUG)
    """: the class specific logger"""

    def __init__(
        self,
        device: str = DEF_DEVICE,
        mjpeg: bool = True,
        port: int = 12345,
    ) -> None:
        """Constructor"""
        self.mjpeg = mjpeg
        self.running = True
        self.clients = []
        self.client_threads = []
        self.device = device
        self.post_process = lambda x: x
        self.port = port

    def stop(self) -> None:
        """Stop thread"""
        self.running = False

    def nr_clients(self) -> int:
        """Number of connected clients"""
        return len(self.clients)

    def _new_client(self, client_socket: socket, addr) -> None:
        """Add new client to list and start its streaming thread"""
        self.log.info(f"{addr} connected.")
        self.clients.append(client_socket)
        self.log.info(f"Number of TCP clients: {self.nr_clients()}")

        tcp_client_thread = threading.Thread(target=self._tcp_thread(client_socket))
        self.client_threads.append(tcp_client_thread)
        self.log.debug(f"Created thread: {tcp_client_thread.name}")
        tcp_client_thread.start()

    def _lost_client(self, client_socket: socket) -> None:
        """Remove client from list"""
        self.log.info(f"TCP Client {client_socket.getsockname()} disconnected")
        client_socket.close()
        self.clients.remove(client_socket)
        self.log.info(f"Number of TCP clients: {self.nr_clients()}")

    def join_all_threads(self) -> None:
        """Join streaming threads from all TCP clients"""
        for t in self.client_threads:
            if t.is_alive():
                self.log.debug(f"Joining: {t.name}")
                t.join()

    def _cleanup_thread_list(self) -> None:
        """Remove inactive threads from list"""
        for t in list(self.client_threads):
            if not t.is_alive():
                self.log.debug(f"Removing dead thread: {t.name}")
                self.client_threads.remove(t)

    def socket_thread(self) -> None:
        """Thread to listen for new connections"""
        try:
            self.log.info(f"Creating TCP socket on port={self.port}")
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.settimeout(1)

            s.bind(
                ("", self.port)
            )  # if the clients/server are on different network you shall bind to ('', port)
            s.listen(10)

            while self.running:
                try:
                    c_socket, addr = s.accept()
                    self._new_client(c_socket, addr)
                except socket.timeout:
                    self._cleanup_thread_list()
                    continue
        except Exception as e:
            self.log.error(f"Exception caught in socket thread: {e}")
            self.stop()
            # sys.exit(0)

    @staticmethod
    def package_mjpeg(img_bytes: bytes, metadata: FrameMetadata = None) -> bytes:
        """Package jpeg bytes with MJPEG headers and frame separator

        Also append metadata if present
        """

        meta = b""
        if metadata:
            meta += b"Timestamp:" + bytes(str(metadata.time), "utf-8")
            meta += b"\r\n"
            meta += b"Sequence:" + bytes(str(metadata.sequence), "utf-8")
            meta += b"\r\n"

        mjpeg = (
            b"Content-Type: image/jpeg\r\nContent-Length: "
            + str(len(meta) + len(img_bytes)).encode("utf-8")
            + b"\r\n\r\n"
        )
        if img_bytes:
            mjpeg += img_bytes + b"\r\n"

        mjpeg += meta

        mjpeg += b"--frame\r\n"

        return mjpeg

    def _tcp_thread(self, client_socket: socket):
        """Streaming thread

        One per client
        """
        # cam = Camera(self.device, stream=True)
        frames_iter = Camera.frames(self.device)
        # Markers for designating the start and end of the info
        # bytes
        marker = b"STREAM"
        marker_end = b"END"

        if self.mjpeg:
            try:
                client_socket.send(b"--frame\r\n")
            except OSError:
                self._lost_client(client_socket)
                return
        else:
            frame, error = next(frames_iter)
            frame = self.post_process(frame)

            # Pack frame specification as bytes
            height = frame.shape[0].to_bytes(length=2, byteorder="big")
            width = frame.shape[1].to_bytes(length=2, byteorder="big")
            match frame.pix_fmt:
                case "RGB3" | "BGR3" | "RGB6":
                    depth = 3
                case "RGB4" | "BGR4":
                    depth = 4
                case _:
                    depth = 1
            depth = depth.to_bytes(length=2, byteorder="big")
            match frame.pix_fmt:
                case "Y16" | "RGB6":
                    n_bytes = (2).to_bytes(length=1)
                case _:
                    n_bytes = (1).to_bytes(length=1)

            msg = marker + height + width + depth + n_bytes + marker_end
            client_socket.sendall(msg)

        for frame, error in frames_iter:
            try:
                if self.mjpeg:
                    jpeg = frame.jpeg()
                    mjpeg = TCPSocket.package_mjpeg(jpeg.data, jpeg.metadata)
                    client_socket.send(mjpeg)
                else:
                    msg = (
                        b"F"
                        + (8).to_bytes(length=2)
                        + frame.metadata.sequence.to_bytes(length=4)
                        + b"E"
                    )
                    client_socket.send(msg)
                    client_socket.send(frame.data)
            except OSError:
                self._lost_client(client_socket)
                break
            except Exception as e:
                self.log.error(f"Exception caught in tcp thread: {e}")
                break

        # while self.running:
        #     try:
        #         if self.mjpeg:
        #             jpeg = cam.get_frame_jpeg()
        #             mjpeg = TCPSocket.package_mjpeg(jpeg.data, jpeg.metadata)
        #             client_socket.send(mjpeg)
        #         else:
        #             frame = cam.get_frame()
        #             client_socket.send(frame.data)
        #     except OSError:
        #         self._lost_client(client_socket)
        #         break
        #     except Exception as e:
        #         self.log.error(f"Exception caught in tcp thread: {e}")
        #         break
