"""Helper functions."""

import importlib.util
import logging
import os
from typing import Optional


try:
    import colorlog

    HAVE_COLORLOG = True
except ImportError:
    HAVE_COLORLOG = False


class ModuleNotInstalledError(ValueError):
    def __init__(self, msg, module_name):
        self.module_name = module_name
        super().__init__(msg)


class DeferredImport:
    def __init__(self, name, error_message=None):
        is_missing = importlib.util.find_spec(name) is None
        self.name = name
        self.msg = (
            f"Library {name} was requested but is not installed"
            if error_message is None
            else error_message
        )
        self.mod = None if is_missing else importlib.import_module("cv2")

    def __getattr__(self, name):
        if self.mod is None:
            raise ModuleNotInstalledError(self.msg, self.name)
        else:
            return self.mod.__getattribute__(name)


def _add_stream_handler(
    logger: logging.Logger,
    level: int = logging.DEBUG,
    color: bool = True,
) -> None:
    """Add a StreamHandler() to the specified logger

    :param logger: logger
    :param level: desired log level, defaults to logging.DEBUG
    :param color: optional coloring, defaults to True (requires 'colorlog')
    """
    c_handler = logging.StreamHandler()
    c_handler.setLevel(level)
    # c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    format_str = (
        "%(asctime)s - %(name)s %(levelname)-8s (%(threadName)-9s) - %(message)s"
    )
    # FORMAT = "[%(asctime)s %(filename)s->%(funcName)s():%(lineno)s]%(levelname)s: %(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"
    if color and HAVE_COLORLOG and os.isatty(2):
        cformat = "%(log_color)s" + format_str
        colors = {
            "DEBUG": "bold_cyan",
            "INFO": "bold_green",
            "WARNING": "bold_yellow",
            "ERROR": "bold_red",
            "CRITICAL": "bold_red",
        }
        c_format = colorlog.ColoredFormatter(cformat, date_format, log_colors=colors)
    else:
        c_format = logging.Formatter(format_str, date_format)
    c_handler.setFormatter(c_format)

    # print(f"adding {c_handler}")
    logger.addHandler(c_handler)


def _add_file_handler(
    logger: logging.Logger,
    level: int = logging.DEBUG,
    log_file: str = "file.log",
) -> None:
    """Add a FileHandler() to the specified logger

    :param logger: logger
    :param level: desired log level, defaults to logging.DEBUG
    :param log_file: path to log file, defaults to 'file.log'
    """
    f_handler = logging.FileHandler(log_file)
    f_handler.setLevel(level)
    f_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    f_handler.setFormatter(f_format)

    # print(f"adding {f_handler}")
    logger.addHandler(f_handler)


def logger_conf(
    name: Optional[str] = None,
    level: int = logging.DEBUG,
    c_level: Optional[int] = None,
    f_level: Optional[int] = None,
    color: bool = True,
    log_file: str = "file.log",
) -> logging.Logger:
    """Setup the logging environment.

    Replaces any existing File and/or Stream Handlers on the logger
    with the specifed ones.

    :param name: logger name, defaults to None (root logger)
    :param level: log level for logger, defaults to logging.DEBUG
    :param c_level: log level for StreamHandler, defaults to None (no handler)
    :param f_level: log level for FileHandler, defaults to None (no handler)
    :param color: optional coloring on StreamHandler, defaults to True
    :param log_file: log file for FileHandler, defaults to "file.log"
    :return: logger
    """
    if c_level and c_level < level:
        level = c_level
    if f_level and f_level < level:
        level = f_level

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Handlers

    # remove existing file/stream handlers
    for h in logger.handlers:
        if type(h) is logging.StreamHandler or type(h) is logging.FileHandler:
            # print(f"removing {h}")
            logger.removeHandler(h)

    if c_level:
        _add_stream_handler(logger, c_level, color)

    if f_level:
        _add_file_handler(logger, f_level, log_file)

    return logger


def get_video_device(link: str) -> str:
    """Get the 'real' video device from a possible video device link

    Fx: '/dev/qtec/video0' -> '/dev/videoX'

    :return: 'real' video device pointed by link (itself if not a link)
    """
    return os.path.realpath(link)


def get_video_device_str(link: str) -> str:
    """Get video device and it's link

    :return: String formated as '{video_dev} -> {link}'
             or just '{video_dev}' if not a link
    """
    dev_path = os.path.realpath(link)
    if dev_path != link:
        return f"{link} -> {dev_path}"
    else:
        return dev_path
